import sys
import kivy
import pygame

kivy.require('1.6.0')

from kivy.app import App
from kivy.core.window import Window
from kivy.uix.widget import Widget
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.scrollview import ScrollView
from kivy.graphics import Color, Rectangle
from kivy.utils import platform

from erdslangetjie.level import LevelList
from erdslangetjie.player import ThePlayer, Nemesis

if platform() != 'android':
    TILE_SIZE = 40
else:
    TILE_SIZE = 64

class GameWindow(RelativeLayout):

    def __init__(self, level_list, view):
        self.level_list = level_list
        self.level_obj = self.level_list.get_current_level()
        self.level_obj.load_tiles()
        self.tiles = {}
        self.view = view

        cols, rows = self.level_obj.get_size()

        super(GameWindow, self).__init__(
                size=(cols * TILE_SIZE, rows * TILE_SIZE),
                size_hint=(None, None))

        self.x_scroll_margin = float(TILE_SIZE) / self.view.size[0]
        self.y_scroll_margin = float(TILE_SIZE) / self.view.size[1]

        self.mouse_move = False

        self.player = ThePlayer()
        self.nemesis = Nemesis()
        if not self.level_obj.enter_pos:
            raise RuntimeError('No entry point')
        self.player_tile = None
        self.nemesis_tile = None

        self.player.pos = self.level_obj.enter_pos
        if platform() != 'android':
            # Very hack'ish
            self.keyboard = Window.request_keyboard(self._closed, self)
            self.keyboard.bind(on_key_down=self._on_key_down)

    def build(self):
        self.clear_widgets()
        self.tiles = {}
        tiles = self.level_obj.get_tiles()
        bx, by = 0, 0
        for tile_line in tiles:
            bx = 0
            for tile in tile_line:
                node = Widget(size=(TILE_SIZE, TILE_SIZE),
                        pos=(bx, by),
                        size_hint=(None, None))
                self.add_widget(node)
                with node.canvas:
                    Color(1, 1, 1)
                    Rectangle(pos=node.pos, size=node.size,
                            texture=tile.texture)
                self.tiles[(bx, by)] = node
                bx += TILE_SIZE
            by += TILE_SIZE

    def draw_player(self):
        if self.player_tile:
            self.remove_widget(self.player_tile)
        sprite_pos = (self.player.pos[0] * TILE_SIZE,
                self.player.pos[1] * TILE_SIZE)
        self.player_tile = Widget(size=(TILE_SIZE, TILE_SIZE),
                pos=sprite_pos)
        with self.player_tile.canvas:
            Color(1, 1, 1)
            Rectangle(pos=sprite_pos, size=self.player_tile.size,
                    texture=self.player.get_texture())
        self.add_widget(self.player_tile)
        for offset in [(TILE_SIZE - 1, TILE_SIZE - 1),
                (-TILE_SIZE + 1, TILE_SIZE - 1),
                (TILE_SIZE - 1, -TILE_SIZE + 1),
                (-TILE_SIZE + 1, -TILE_SIZE + 1),
                (0, 2 * TILE_SIZE - 2),
                (-2 * TILE_SIZE + 2, 0),
                (2 * TILE_SIZE - 2, 0),
                (0, -2 * TILE_SIZE + 2),
                (0, 0)]:
            # Aim is to ensure a 'neighbourhood' around the player
            # is visible if possible
            check_point = (sprite_pos[0] + offset[0] + TILE_SIZE / 2,
                    sprite_pos[1] + offset[1] + TILE_SIZE / 2)
            true_point = self.to_parent(*check_point)
            if check_point[0] < 0:
                continue
            if check_point[1] < 0:
                continue
            if check_point[0] >= self.size[0]:
                continue
            if check_point[1] >= self.size[1]:
                continue
            while not self.included(true_point, 0):
                # Scroll ourselves
                if true_point[0] >= self.view.size[0]:
                    self.view.scroll_x += self.x_scroll_margin
                    true_point = self.to_parent(*check_point)
                    #print '-x', self.view.scroll_x, self.view.scroll_y
                elif true_point[0] < 0:
                    self.view.scroll_x -= self.x_scroll_margin
                    true_point = self.to_parent(*check_point)
                    #print '+x', self.view.scroll_x, self.view.scroll_y
                elif true_point[1] >= self.view.size[1]:
                    self.view.scroll_y += self.y_scroll_margin
                    true_point = self.to_parent(*check_point)
                    #print '+y', self.view.scroll_x, self.view.scroll_y
                elif true_point[1] < 0:
                    self.view.scroll_y -= self.y_scroll_margin
                    true_point = self.to_parent(*check_point)
                    #print '-y', self.view.scroll_x, self.view.scroll_y
                #print true_point, self.view.size

    def included(self, point, margin):
        if point[0] < margin:
            return False
        if point[0] >= self.view.size[0] - margin:
            return False
        if point[1] < margin:
            return False
        if point[1] >= self.view.size[1] - margin:
            return False
        return True

    def draw_nemesis(self):
        if not self.nemesis.on_board():
            return
        if self.nemesis_tile:
            self.remove_widget(self.nemesis_tile)
        sprite_pos = (self.nemesis.pos[0] * TILE_SIZE,
                self.nemesis.pos[1] * TILE_SIZE)
        self.nemesis_tile = Widget(size=(TILE_SIZE, TILE_SIZE),
                pos=sprite_pos)
        with self.nemesis_tile.canvas:
            Color(1, 1, 1)
            Rectangle(pos=sprite_pos, size=self.nemesis_tile.size,
                    texture=self.nemesis.get_texture())
        self.add_widget(self.nemesis_tile)

    def _closed(self):
        self.keyboard.unbind(on_key_down=self._on_key_down)

    def _on_key_down(self, keyboard, keycode, text, modifiers):
        # FIXME - likely portablity issues
        direction = None
        if keycode[0] == pygame.K_UP:
            direction = (0, 1)
        elif keycode[0] == pygame.K_DOWN:
            direction = (0, -1)
        elif keycode[0] == pygame.K_LEFT:
            direction = (-1, 0)
        elif keycode[0] == pygame.K_RIGHT:
            direction = (1, 0)
        if direction:
            self.do_move(direction)

    def do_move(self, direction):
        self.nemesis.move(self.level_obj)
        self.draw_nemesis()
        self.player.move(direction, self.level_obj)
        self.draw_player()
        self.check_state()

    def check_state(self):
        if self.level_obj.at_exit(self.player.pos):
            # Jump to next level
            self.level_obj = self.level_list.advance_to_next_level()
            self.remove_widget(self.nemesis_tile)
            self.nemesis.reset_pos()
            if self.level_obj:
                self.level_obj.load_tiles()
                self.player.pos = self.level_obj.enter_pos
                self.remove_widget(self.player_tile)
                self.view.scroll_x = 0
                self.view.scroll_y = 0
                self.build()
                self.draw_nemesis()
                self.draw_player()
            else:
                print 'You won!'
                sys.exit(1)
        elif self.nemesis.pos == self.player.pos:
            # Caught
            print 'You lost!'
            sys.exit(1)

    def _calc_mouse_pos(self, pos):
        pos = self.to_local(*pos)
        return (int(pos[0] / TILE_SIZE), int(pos[1] / TILE_SIZE))

    def on_touch_down(self, touch):
        pos = self._calc_mouse_pos(touch.pos)
        if pos == self.player.pos:
            self.mouse_move = True
            self.mouse_start = pos

    def on_touch_up(self, touch):
        self.mouse_move = False

    def on_touch_move(self, touch):
        if self.mouse_move:
            pos = self._calc_mouse_pos(touch.pos)
            if (pos[0] - self.mouse_start[0] != 0) or (
                    pos[1] - self.mouse_start[1] != 0):
                direction = (pos[0] - self.mouse_start[0],
                        pos[1] - self.mouse_start[1])
                self.do_move(direction)
                self.mouse_start = pos


class GameApp(App):

    title = "Peter's thread snake"

    def __init__(self):
        self.levels = LevelList()
        super(GameApp, self).__init__()

    def build(self):
        root = ScrollView(size_hint=(None, None))
        return root

    def on_start(self):
        from kivy.base import EventLoop
        window = EventLoop.window
        if platform() == 'android':
            window.fullscreen = True
        self.root.size = window.size
        game = GameWindow(self.levels, self.root)
        game.build()
        self.root.add_widget(game)
        # Ensure the player is visible
        self.root.scroll_x = 0
        self.root.scroll_y = 0
        game.draw_player()
        game.draw_nemesis()


def main():
    """ Erdslangetjie, a maze game of eluding nemesis
    """
    GameApp().run()
